import argparse
import ast
from maldi_nn.models import MaldiTransformer
from maldi_nn.utils.data import DRIAMSSpectrumDataModule
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch import Trainer
from maldi_nn.spectrum import PeakFilter
from maldi_nn.reproduce.modules import (
    NegativeSamplerPlugin,
    MaskingPlugin,
    DRIAMSSpectrumDataModuleWithNoiser,
    MaldiTransformerNegSampler,
    MaldiTransformerOnlyClf,
    MaldiTransformerMaskMSE
)
import h5torch

def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def main():

    class CustomFormatter(
        argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
    ):
        pass

    parser = argparse.ArgumentParser(
        description="Training script for Maldi Transformer.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument("logs_path", type=str, metavar="logs_path", help="path to logs.")
    parser.add_argument(
        "spectrum_embedder",
        type=str,
        metavar="spectrum_embedder",
        choices=["S", "M", "L", "XL"],
        help="Which size spectrum embedder to use",
    )

    parser.add_argument(
        "--mode",
        type=str,
        default="vanilla",
        choices=["vanilla", "negpeaksampler", "intensitymlm", "onlyclf", "onlyshf"],
        help="Maldi Transformer training mode, choices: {%(choices)s} Note that negpeaksampler requires to run reproduce.estimate_peak_distr first.",
        )

    parser.add_argument(
        "--n_peaks",
        type=int,
        default=200,
        help="Number of peaks",
    )

    parser.add_argument(
        "--p",
        type=float,
        default=0.15,
        help="shuffle freq",
    )

    parser.add_argument(
        "--lmbda",
        type=float,
        default=1 / 100,
        help="Lambda. This is the probability with which to apply the spec id loss per step. ",
    )

    parser.add_argument(
        "--lmbda2",
        type=float,
        default=1.,
        help="Additionally, fixed multiplier to apply to the spec id loss.",
    )

    parser.add_argument(
        "--steps",
        type=float,
        default=500_000,
        help="steps",
    )

    parser.add_argument(
        "--batch_size",
        type=int,
        default=512,
        help="batch size per gpu, effective batch size is this value times the number of gpus. (pytorch ddp).",
    )

    parser.add_argument("--lr", type=float, default=5e-4, help="Learning rate.")

    parser.add_argument(
        "--num_workers",
        type=int,
        default=4,
        help="Number of workers in dataloader. Reduce to alleviate CPU.",
    )
    parser.add_argument(
        "--devices",
        type=ast.literal_eval,
        default=1,
        help="devices to use. Input an integer to specify a number of gpus or a list e.g. [1] or [0,1,3] to specify which gpus.",
    )

    args = parser.parse_args()

    size_to_layer_dims = {
        "S": [160, 4],
        "M": [184, 6],
        "L": [232, 8],
        "XL": [304, 10],
    }

    if args.mode == "negpeaksampler":
        dm = DRIAMSSpectrumDataModuleWithNoiser(
            args.path,
            batch_size=args.batch_size,
            n_workers=args.num_workers,
            preprocessor=PeakFilter(max_number=args.n_peaks),
            min_spectrum_len=128,
            in_memory=True,
            exclude_nans=False,
            noiser=NegativeSamplerPlugin(
                p_x=h5torch.File(args.path)["unstructured/p_x_200"][:],
                p_y_x=h5torch.File(args.path)["unstructured/p_y_x_200"][:],
                train_on=int(args.n_peaks * args.p / 2),
                n_to_sample=int(args.n_peaks * args.p / 2),
                range=0.5,
                binningstep=1,
            ),
        )
    elif args.mode == "intensitymlm":
        dm = DRIAMSSpectrumDataModuleWithNoiser(
            args.path,
            batch_size=args.batch_size,
            n_workers=args.num_workers,
            preprocessor=PeakFilter(max_number=args.n_peaks),
            min_spectrum_len=128,
            in_memory=True,
            exclude_nans=False,
            noiser=MaskingPlugin(prob=args.p, unchanged=0.2),
        )
    else:
        dm = DRIAMSSpectrumDataModule(
            args.path,
            batch_size=args.batch_size,
            n_workers=args.num_workers,
            preprocessor=PeakFilter(max_number=args.n_peaks),
            min_spectrum_len=128,
            in_memory=True,
            exclude_nans=(True if args.mode == "onlyclf" else False),
        )


    modeltype = MaldiTransformer
    model_kwargs = {
        "n_classes" : dm.n_species - 1,
        "n_heads" : 8,
        "dropout" : 0.2,
        "p" : args.p,
        "clf" : True,
        "clf_train_p" : args.lmbda,
        "lmbda" : args.lmbda2,
        "lr" : args.lr,
        "weight_decay" : 0,
        "lr_decay_factor" : 1,
        "warmup_steps" : 2500,
    }

    if args.mode == "onlyclf":
        modeltype = MaldiTransformerOnlyClf
        model_kwargs["n_classes"] = dm.n_species
        del model_kwargs["clf"]
        del model_kwargs["p"]
        del model_kwargs["clf_train_p"]
        del model_kwargs["lmbda"]
    elif args.mode == "negpeaksampler":
        modeltype = MaldiTransformerNegSampler
        del model_kwargs["lmbda"]
        del model_kwargs["p"]
    elif args.mode == "intensitymlm":
        modeltype = MaldiTransformerMaskMSE
        del model_kwargs["lmbda"]
        del model_kwargs["p"]
    elif args.mode == "onlyshf":
        model_kwargs["clf"] = False


    model = modeltype(
        size_to_layer_dims[args.spectrum_embedder][1],
        size_to_layer_dims[args.spectrum_embedder][0],
        **model_kwargs,
    )


    if args.mode == "onlyclf":
        val_ckpt = ModelCheckpoint(monitor="val_clfacc", mode="max")
        callbacks = [val_ckpt, EarlyStopping(monitor="val_clfacc", patience=10, mode="max")]
        logger = TensorBoardLogger(
            args.logs_path,
            name="malditrf%s_%s_%s_%s"
            % (args.mode, args.spectrum_embedder, args.n_peaks, args.lr),
        )
        trainer = Trainer(
            accelerator="gpu",
            devices=args.devices,
            strategy="auto",
            max_epochs=250,
            callbacks=callbacks,
            logger=logger,
            precision="bf16-mixed",
        )

    else:
        logger = TensorBoardLogger(
            args.logs_path,
            name="malditrf%s_%s_%s_%s_%s_%s"
            % (args.mode, args.spectrum_embedder, args.n_peaks, args.p, args.lmbda, args.lr),
        )
        callbacks = [
            ModelCheckpoint(every_n_train_steps=10_000),
        ]

        trainer = Trainer(
            accelerator="gpu",
            devices=args.devices,
            strategy="auto",
            gradient_clip_val=1,
            max_steps=args.steps,
            max_epochs=-1,
            val_check_interval=5_000,
            check_val_every_n_epoch=None,
            callbacks=callbacks,
            logger=logger,
            precision="bf16-mixed",
        )

    trainer.fit(model, dm)


if __name__ == "__main__":
    main()


from maldi_nn.utils.data import *
import sys
from importlib.resources import files
import h5torch
import json
import os
import pandas as pd
import numpy as np
from maldi_nn.spectrum import *
import shutil

def RKI_raw_to_h5torch(RKI_ROOT, outfile):
    data_path = files("maldi_nn.utils").joinpath("RKI_split.json")
    split = json.load(open(data_path))

    species = []
    subspecies = []
    spectra = []
    loc = []
    for l in os.walk(RKI_ROOT):
        root, subfolders, files_ = l
        if "fid" in files_ and "acqu" in files_:
            r_ = root.split("/")
            species.append(r_[-6])
            subspecies.append(r_[-5])
            loc.append(root.lstrip(RKI_ROOT))
            s = SpectrumObject.from_bruker(
                os.path.join(root, "acqu"), os.path.join(root, "fid")
            )
            spectra.append(s)

    species_map = {k: v for v, k in enumerate(np.unique(species))}
    species_map_inv = {v: k for k, v in species_map.items()}
    species_labels = np.array([species_map_inv[i] for i in range(len(species_map_inv))])
    species = np.array([species_map[i] for i in species])

    spectrum_split = np.array([split[l] for l in loc])

    f = h5torch.File(outfile, "w")
    f.register(species, "central")
    f.register(np.array(loc).astype(bytes), 0, name="loc")
    f.register(species_labels.astype(bytes), "unstructured", name="species_labels")
    f.register(spectrum_split.astype(bytes), "unstructured", name="split")

    ints = [s.intensity for s in spectra]
    mzs = [s.mz for s in spectra]

    f.register(ints, 0, name="intensity", mode="vlen")
    f.register(mzs, 0, name="mz", mode="vlen")
    f.close()


def RKI_raw_to_binned(rawfile, processed_file):
    binner = SequentialPreprocessor(
        VarStabilizer(method="sqrt"),
        Smoother(halfwindow=10),
        BaselineCorrecter(method="SNIP", snip_n_iter=20),
        Trimmer(),
        Binner(),
        Normalizer(sum=1),
    )
    shutil.copy(rawfile, processed_file)
    file = h5torch.File(processed_file, "a")
    len_ = file["0/mz"].shape[0]
    ints = []
    for i in range(len_):
        mz = file["0/mz"][i]
        intensity = file["0/intensity"][i]
        s = SpectrumObject(mz=mz, intensity=intensity)
        ints.append(binner(s).intensity)
        if (i + 1) % 1000 == 0:
            print(i, end=" ", flush=True)

    del file["0/mz"]
    del file["0/intensity"]

    file.register(np.stack(ints), 0, name="intensity")
    file.register(binner(s).mz, "unstructured", name="mz")

    file.close()
    return None


def RKI_raw_to_peaks(rawfile, processed_file):
    peakdetector = SequentialPreprocessor(
        VarStabilizer(method="sqrt"),
        Smoother(halfwindow=10),
        BaselineCorrecter(method="SNIP", snip_n_iter=20),
        Trimmer(),
        PersistenceTransformer(extract_nonzero=True),
        Normalizer(sum=1),
        PeakFilter(max_number=2048),
    )
    shutil.copy(rawfile, processed_file)
    file = h5torch.File(processed_file, "a")
    len_ = file["0/mz"].shape[0]
    ints = []
    mzs = []
    for i in range(len_):
        mz = file["0/mz"][i]
        intensity = file["0/intensity"][i]
        s = SpectrumObject(mz=mz, intensity=intensity)
        s = peakdetector(s)
        ints.append(s.intensity)
        mzs.append(s.mz)
        if (i + 1) % 1000 == 0:
            print(i, end=" ", flush=True)

    del file["0/mz"]
    del file["0/intensity"]
    file.register(ints, 0, name="intensity", mode="vlen")
    file.register(mzs, 0, name="mz", mode="vlen")

    file.close()
    return None


def main():
    RKI_root = str(sys.argv[1])
    spectraraw = str(sys.argv[2])
    spectrabin = str(sys.argv[3])
    spectrapks = str(sys.argv[4])

    RKI_raw_to_h5torch(RKI_root, spectraraw)
    RKI_raw_to_binned(spectraraw, spectrabin)
    RKI_raw_to_peaks(spectraraw, spectrapks)

if __name__ == "__main__":
    main()


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from maldi_nn.models import MaldiLightningModule
from torchmetrics.classification import BinaryAUROC, MulticlassAccuracy
import maldi_nn.nn as maldinn
from maldi_nn.utils.data import DRIAMSSpectrumDataModule
from maldi_nn.spectrum import (
    Binner,
    SpectrumObject,
    UniformPeakShifter,
    SequentialPreprocessor,
)


class MaldiTransformerOnlyClf(MaldiLightningModule):
    def __init__(
        self,
        depth,
        dim,
        n_classes=64,
        n_heads=8,
        dropout=0.2,
        lr=0.0005,
        weight_decay=0,
        lr_decay_factor=1,
        warmup_steps=2500,
    ):
        super().__init__(
            lr=lr,
            weight_decay=weight_decay,
            lr_decay_factor=lr_decay_factor,
            warmup_steps=warmup_steps,
        )
        self.save_hyperparameters()
        self.transformer = maldinn.Transformer(
            depth,
            dim,
            n_heads=n_heads,
            dropout=dropout,
            cls=True,
            reduce="none",
            output_head_dim=n_classes,
        )

        self.n_species = n_classes

        self.auroc = BinaryAUROC()
        self.accuracy = MulticlassAccuracy(num_classes=n_classes, average="micro")
        self.top5_accuracy = MulticlassAccuracy(
            num_classes=n_classes, top_k=5, average="micro"
        )

    def forward(self, batch):
        z = self.transformer(batch)
        clf_logits = self.transformer.output_head(z[:, 0, :])
        return clf_logits

    def training_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        clf_logits = self(batch)

        clf_loss = F.cross_entropy(clf_logits, batch["species"])

        self.log("train_loss", clf_loss, batch_size=len(batch["species"]))
        return clf_loss

    def validation_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        clf_logits = self(batch)

        clf_loss = F.cross_entropy(clf_logits, batch["species"])

        self.log(
            "val_clfloss",
            clf_loss,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["species"]),
            sync_dist=True,
        )
        self.accuracy(clf_logits, batch["species"])
        self.log(
            "val_clfacc",
            self.accuracy,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["species"]),
            sync_dist=True,
        )
        self.top5_accuracy(clf_logits, batch["species"])
        self.log(
            "val_clftop5_acc",
            self.top5_accuracy,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["species"]),
            sync_dist=True,
        )

    def predict_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        clf_logits = self(batch)

        return (
            batch["loc"],
            clf_logits,
            batch["species"],
        )


class MaskingPlugin:
    def __init__(self, prob=0.15, unchanged=0.2):
        self.p = prob
        self.u = unchanged

    def __call__(self, spectrum):
        spectrum["intensity_true"] = spectrum["intensity"].clone()
        train_indices = torch.empty(spectrum["intensity"].shape).uniform_() < self.p

        unchanged_indices = torch.empty(spectrum["intensity"].shape).uniform_() < self.u

        spectrum["intensity"].masked_fill_(
            train_indices & ~unchanged_indices, torch.nan
        )
        spectrum["train_indices"] = train_indices
        return spectrum


class NegativeSamplerPlugin:
    def __init__(
        self, p_x, p_y_x, train_on=15, n_to_sample=15, range=0.5, binningstep=1
    ):
        self.p_x = p_x
        self.p_y_x = p_y_x
        self.train_on = train_on
        self.n = n_to_sample
        self.binner = SequentialPreprocessor(
            Binner(step=binningstep), UniformPeakShifter(range=range)
        )
        self.range_ = torch.arange(self.n)

    def __call__(self, spectrum):
        s = SpectrumObject(
            mz=spectrum["mz"].numpy(), intensity=spectrum["intensity"].numpy()
        )
        s = self.binner(s)

        bg_signal = self.p_x.copy()

        neg_locs = np.random.choice(len(bg_signal), size=(self.n), p=bg_signal)
        xs = s.mz[neg_locs]

        quantiles = self.p_y_x[neg_locs]

        q = torch.randint(100, (self.n,))
        ys = (
            torch.rand(self.n, dtype=spectrum["intensity"].dtype)
            * (quantiles[self.range_, q + 1] - quantiles[self.range_, q])
            + quantiles[self.range_, q]
        )

        train_indices = torch.empty(spectrum["intensity"].shape).uniform_() < (
            self.train_on / len(spectrum["intensity"])
        )

        spectrum["mz"] = torch.cat(
            [spectrum["mz"], torch.tensor(xs, dtype=spectrum["mz"].dtype)]
        )
        spectrum["intensity_true"] = torch.cat(
            [
                torch.ones_like(spectrum["intensity"]).long(),
                torch.zeros((len(neg_locs)), dtype=spectrum["intensity"].dtype),
            ]
        )
        spectrum["intensity"] = torch.cat([spectrum["intensity"], ys])

        spectrum["train_indices"] = torch.cat(
            [train_indices, torch.ones(len(neg_locs), dtype=bool)]
        )
        return spectrum


class DRIAMSSpectrumDataModuleWithNoiser(DRIAMSSpectrumDataModule):
    def __init__(
        self,
        path,
        batch_size=512,
        n_workers=4,
        preprocessor=None,
        min_spectrum_len=None,
        in_memory=True,
        exclude_nans=False,
        noiser=None,
    ):
        super().__init__(
            path=path,
            batch_size=batch_size,
            n_workers=n_workers,
            preprocessor=preprocessor,
            min_spectrum_len=min_spectrum_len,
            in_memory=in_memory,
            exclude_nans=exclude_nans,
        )
        self.noiser = noiser

    def processor(self, f, sample):
        spectrum = SpectrumObject(
            mz=(sample["0/mz"] if "0/mz" in sample else f["unstructured/mz"][:]),
            intensity=sample["0/intensity"],
        )

        if self.preprocessor is not None:
            spectrum = self.preprocessor(spectrum)

        spectrum = {
            "intensity": torch.tensor(spectrum.intensity).float(),
            "mz": torch.tensor(spectrum.mz),
        }

        if self.noiser is not None:
            spectrum = self.noiser(spectrum)

        return (
            spectrum
            | {
                "species": (
                    sample["central"]
                    if self.species_mapping is None
                    else self.species_mapping[sample["central"]]
                )
            }
            | {"loc": sample["0/loc"].astype(str)}
        )


class MaldiTransformerMaskMSE(MaldiLightningModule):
    def __init__(
        self,
        depth,
        dim,
        n_classes=64,
        n_heads=8,
        dropout=0.2,
        clf=True,
        clf_train_p=1 / 100,
        lr=0.0005,
        weight_decay=0,
        lr_decay_factor=1,
        warmup_steps=2500,
    ):
        super().__init__(
            lr=lr,
            weight_decay=weight_decay,
            lr_decay_factor=lr_decay_factor,
            warmup_steps=warmup_steps,
        )
        self.save_hyperparameters()
        self.transformer = maldinn.Transformer(
            depth,
            dim,
            n_heads=n_heads,
            dropout=dropout,
            cls=True,
            reduce="none",
            output_head_dim=n_classes,
        )

        self.output_head = nn.Linear(dim, 1)

        self.n_species = n_classes
        self.clf = clf
        self.clf_train_p = clf_train_p

        self.accuracy = MulticlassAccuracy(num_classes=n_classes, average="micro")
        self.top5_accuracy = MulticlassAccuracy(
            num_classes=n_classes, top_k=5, average="micro"
        )

    def forward(self, batch):
        z = self.transformer(batch)
        mlm_logits = self.output_head(z[:, 1:]).squeeze(-1)
        if self.clf:
            clf_logits = self.transformer.output_head(z[:, 0, :])
            return mlm_logits, clf_logits
        else:
            return mlm_logits

    def training_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        outputs = self(batch)
        if self.clf:
            mlm_logits, clf_logits = outputs
        else:
            mlm_logits = outputs

        mlm_logits_train = self.train_indices_select(mlm_logits, batch["train_indices"])
        trues_train = self.train_indices_select(
            batch["intensity_true"], batch["train_indices"]
        )
        mse_loss = F.mse_loss(mlm_logits_train, trues_train.to(self.dtype))

        if self.clf:
            indexer = batch["species"] < self.n_species
            clf_loss = F.cross_entropy(clf_logits[indexer], batch["species"][indexer])

            self.log(
                "train_loss", mse_loss + clf_loss, batch_size=len(batch["intensity"])
            )
            return mse_loss + clf_loss * (
                1 if (torch.rand(1) < self.clf_train_p).item() else 0
            )
        else:
            self.log("train_loss", mse_loss, batch_size=len(batch["intensity"]))
            return mse_loss

    def validation_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        outputs = self(batch)
        if self.clf:
            mlm_logits, clf_logits = outputs
        else:
            mlm_logits = outputs

        mlm_logits_train = self.train_indices_select(mlm_logits, batch["train_indices"])
        trues_train = self.train_indices_select(
            batch["intensity_true"], batch["train_indices"]
        )
        mse_loss = F.mse_loss(mlm_logits_train, trues_train.to(self.dtype))

        self.log(
            "val_mseloss",
            mse_loss,
            on_step=False,
            on_epoch=True,
            batch_size=len(mlm_logits_train),
            sync_dist=True,
        )

        if self.clf:
            indexer = batch["species"] < self.n_species
            if indexer.sum() > 0:
                clf_loss = F.cross_entropy(
                    clf_logits[indexer], batch["species"][indexer]
                )

                self.log(
                    "val_clfloss",
                    clf_loss,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )
                self.accuracy(clf_logits[indexer], batch["species"][indexer])
                self.log(
                    "val_clfacc",
                    self.accuracy,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )
                self.top5_accuracy(clf_logits[indexer], batch["species"][indexer])
                self.log(
                    "val_clftop5_acc",
                    self.top5_accuracy,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )


class MaldiTransformerNegSampler(MaldiLightningModule):
    def __init__(
        self,
        depth,
        dim,
        n_classes=64,
        n_heads=8,
        dropout=0.2,
        clf=True,
        clf_train_p=1 / 100,
        lr=0.0005,
        weight_decay=0,
        lr_decay_factor=1,
        warmup_steps=2500,
    ):
        super().__init__(
            lr=lr,
            weight_decay=weight_decay,
            lr_decay_factor=lr_decay_factor,
            warmup_steps=warmup_steps,
        )
        self.save_hyperparameters()
        self.transformer = maldinn.Transformer(
            depth,
            dim,
            n_heads=n_heads,
            dropout=dropout,
            cls=True,
            reduce="none",
            output_head_dim=n_classes,
        )

        self.output_head = nn.Linear(dim, 1)

        self.n_species = n_classes
        self.clf = clf
        self.clf_train_p = clf_train_p

        self.auroc = BinaryAUROC()
        self.accuracy = MulticlassAccuracy(num_classes=n_classes, average="micro")
        self.top5_accuracy = MulticlassAccuracy(
            num_classes=n_classes, top_k=5, average="micro"
        )

    def forward(self, batch):
        z = self.transformer(batch)
        mlm_logits = self.output_head(z[:, 1:]).squeeze(-1)
        if self.clf:
            clf_logits = self.transformer.output_head(z[:, 0, :])
            return mlm_logits, clf_logits
        else:
            return mlm_logits

    def training_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        outputs = self(batch)
        if self.clf:
            mlm_logits, clf_logits = outputs
        else:
            mlm_logits = outputs

        mlm_logits_train = self.train_indices_select(mlm_logits, batch["train_indices"])
        trues_train = self.train_indices_select(
            batch["intensity_true"], batch["train_indices"]
        )
        mlm_loss = F.binary_cross_entropy_with_logits(
            mlm_logits_train, trues_train.to(self.dtype)
        )

        if self.clf:
            indexer = batch["species"] < self.n_species
            clf_loss = F.cross_entropy(clf_logits[indexer], batch["species"][indexer])

            self.log(
                "train_loss", mlm_loss + clf_loss, batch_size=len(batch["intensity"])
            )
            return mlm_loss + clf_loss * (
                1 if (torch.rand(1) < self.clf_train_p).item() else 0
            )
        else:
            self.log("train_loss", mlm_loss, batch_size=len(batch["intensity"]))
            return mlm_loss

    def validation_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        outputs = self(batch)
        if self.clf:
            mlm_logits, clf_logits = outputs
        else:
            mlm_logits = outputs

        mlm_logits_train = self.train_indices_select(mlm_logits, batch["train_indices"])
        trues_train = self.train_indices_select(
            batch["intensity_true"], batch["train_indices"]
        )
        mlm_loss = F.binary_cross_entropy_with_logits(
            mlm_logits_train, trues_train.to(self.dtype)
        )

        self.log(
            "val_mlmloss",
            mlm_loss,
            on_step=False,
            on_epoch=True,
            batch_size=len(mlm_logits_train),
            sync_dist=True,
        )

        if self.clf:
            indexer = batch["species"] < self.n_species
            if indexer.sum() > 0:
                clf_loss = F.cross_entropy(
                    clf_logits[indexer], batch["species"][indexer]
                )

                self.log(
                    "val_clfloss",
                    clf_loss,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )
                self.accuracy(clf_logits[indexer], batch["species"][indexer])
                self.log(
                    "val_clfacc",
                    self.accuracy,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )
                self.top5_accuracy(clf_logits[indexer], batch["species"][indexer])
                self.log(
                    "val_clftop5_acc",
                    self.top5_accuracy,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from lightning import LightningModule
from torchmetrics.classification import BinaryAUROC, MulticlassAccuracy
from maldi_nn.utils.drug import (
    DrugOneHotEmbedding,
    DrugMLP,
    DrugGRU,
    DrugCNN,
    DrugTransformer,
    DrugImageCNN,
)
import maldi_nn.nn as maldinn


class MaldiLightningModule(LightningModule):
    def __init__(self, lr=1e-4, weight_decay=0, lr_decay_factor=0.95, warmup_steps=500):
        """
        A wrapper around pl.LightningModule with some standard stuff applicable to all models.
        """
        super().__init__()

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.hparams.lr,
            weight_decay=self.hparams.weight_decay,
        )
        lambd = lambda epoch: self.hparams.lr_decay_factor
        lr_scheduler = torch.optim.lr_scheduler.MultiplicativeLR(
            optimizer, lr_lambda=lambd
        )
        return [optimizer], [lr_scheduler]

    def optimizer_step(self, epoch, batch_idx, optimizer, optimizer_closure=None):
        # warm up lr
        if self.trainer.global_step < self.hparams.warmup_steps:
            lr_scale = min(
                1.0, float(self.trainer.global_step + 1) / self.hparams.warmup_steps
            )
            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.lr

        # update params
        optimizer.step(closure=optimizer_closure)

    def n_params(self):
        params_per_layer = [(name, p.numel()) for name, p in self.named_parameters()]
        total_params = sum(p.numel() for p in self.parameters())
        params_per_layer += [("total", total_params)]
        return params_per_layer

    def train_indices_select(self, in_, train_indices):
        if train_indices.dtype == torch.long:
            return torch.gather(in_, 1, train_indices)
        elif train_indices.dtype == torch.bool:
            return in_[train_indices]
        else:
            raise ValueError("train_indices should be either .long or .bool")


class AMRModel(MaldiLightningModule):
    def __init__(
        self,
        spectrum_embedder="mlp",
        drug_embedder="ecfp",
        spectrum_kwargs={},
        drug_kwargs={},
        lr=1e-4,
        weight_decay=0,
        lr_decay_factor=1.00,
        warmup_steps=250,
        scaled_dot_product=False,
    ):
        super().__init__(
            lr=lr,
            weight_decay=weight_decay,
            lr_decay_factor=lr_decay_factor,
            warmup_steps=warmup_steps,
        )
        self.save_hyperparameters()

        if drug_embedder == "onehot":
            self.drug_embedder = DrugOneHotEmbedding(**drug_kwargs)
        elif drug_embedder == "ecfp":
            self.drug_embedder = DrugMLP(**drug_kwargs)
        elif drug_embedder == "gru":
            self.drug_embedder = DrugGRU(**drug_kwargs)
        elif drug_embedder == "cnn":
            self.drug_embedder = DrugCNN(**drug_kwargs)
        elif drug_embedder == "trf":
            self.drug_embedder = DrugTransformer(**drug_kwargs)
        elif drug_embedder == "img":
            self.drug_embedder = DrugImageCNN(**drug_kwargs)
        elif drug_embedder == "kernel":
            self.drug_embedder = DrugMLP(**drug_kwargs)

        if spectrum_embedder == "mlp":
            self.spectrum_embedder = maldinn.MLP(**spectrum_kwargs)
        elif spectrum_embedder == "trf":
            self.spectrum_embedder = maldinn.Transformer(**spectrum_kwargs)

        self.scale = scaled_dot_product

        self.auroc = BinaryAUROC()

    def forward(self, batch):
        drug_embedding = self.drug_embedder(batch["drug"])
        spectrum_embedding = self.spectrum_embedder(batch)
        norm = 1 if not self.scale else spectrum_embedding.shape[-1] ** 0.5
        return (drug_embedding * spectrum_embedding).sum(-1) / norm

    def training_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)
        batch["drug"] = batch["drug"].to(self.dtype)

        logits = self(batch)

        loss = F.binary_cross_entropy_with_logits(
            logits, batch["label"].to(logits.dtype)
        )

        self.log("train_loss", loss, batch_size=len(batch["label"]))
        return loss

    def validation_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)
        batch["drug"] = batch["drug"].to(self.dtype)

        logits = self(batch)

        loss = F.binary_cross_entropy_with_logits(
            logits, batch["label"].to(logits.dtype)
        )

        self.log(
            "val_loss",
            loss,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["label"]),
        )

        self.auroc(logits, batch["label"])
        self.log(
            "val_micro_auc",
            self.auroc,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["label"]),
        )

    def predict_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)
        batch["drug"] = batch["drug"].to(self.dtype)

        logits = self(batch)

        return (
            torch.stack([batch["label"].to(logits), logits], -1),
            batch["loc"],
            batch["drug_name"],
        )


class SpeciesClassifier(MaldiLightningModule):
    def __init__(
        self,
        spectrum_embedder="mlp",
        spectrum_kwargs={},
        lr=1e-4,
        weight_decay=0,
        lr_decay_factor=1.00,
        warmup_steps=250,
        genus_labels = None,
    ):
        super().__init__(
            lr=lr,
            weight_decay=weight_decay,
            lr_decay_factor=lr_decay_factor,
            warmup_steps=warmup_steps,
        )
        self.save_hyperparameters()

        if spectrum_embedder == "mlp":
            self.spectrum_embedder = maldinn.MLP(**spectrum_kwargs)
            n_classes = spectrum_kwargs["n_outputs"]
        elif spectrum_embedder == "trf":
            self.spectrum_embedder = maldinn.Transformer(**spectrum_kwargs)
            n_classes = spectrum_kwargs["output_head_dim"]

        self.accuracy = MulticlassAccuracy(num_classes=n_classes, average="micro")

        if genus_labels is not None:
            self.genus_accuracy = MulticlassAccuracy(num_classes=torch.unique(genus_labels).shape[0], average="micro")
            self.genus_labels = genus_labels

    def forward(self, batch):
        return self.spectrum_embedder(batch)

    def training_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        logits = self(batch)

        loss = F.cross_entropy(logits, batch["species"])

        self.log("train_loss", loss, batch_size=len(batch["species"]))
        return loss

    def validation_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        logits = self(batch)

        loss = F.cross_entropy(logits, batch["species"])

        self.log(
            "val_loss",
            loss,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["species"]),
        )

        self.accuracy(logits, batch["species"])
        self.log(
            "val_acc",
            self.accuracy,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["species"]),
        )
        self.genus_accuracy(self.genus_labels.to(logits.device)[logits.argmax(-1)], self.genus_labels.to(logits.device)[batch["species"]])
        self.log(
            "val_genus_acc",
            self.genus_accuracy,
            on_step=False,
            on_epoch=True,
            batch_size=len(batch["species"]),
        )

    def predict_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        logits = self(batch)

        return (
            torch.cat([batch["species"].to(logits).unsqueeze(-1), logits], -1),
            batch["0/loc"],
        )


class MaldiTransformer(MaldiLightningModule):
    def __init__(
        self,
        depth,
        dim,
        n_classes=64,
        n_heads=8,
        dropout=0.2,
        p=0.125,
        clf=False,
        clf_train_p=1 / 100,
        lr=0.0005,
        weight_decay=0,
        lr_decay_factor=1,
        warmup_steps=2500,
        lmbda = 1,
    ):
        super().__init__(
            lr=lr,
            weight_decay=weight_decay,
            lr_decay_factor=lr_decay_factor,
            warmup_steps=warmup_steps,
        )
        self.save_hyperparameters()
        self.transformer = maldinn.Transformer(
            depth,
            dim,
            n_heads=n_heads,
            dropout=dropout,
            cls=True,
            reduce="none",
            output_head_dim=n_classes,
        )

        self.output_head = nn.Linear(dim, 1)

        self.n_species = n_classes
        self.clf = clf
        self.clf_train_p = clf_train_p
        self.p = p
        self.lmbda = lmbda

        self.auroc = BinaryAUROC()
        self.accuracy = MulticlassAccuracy(num_classes=n_classes, average="micro")
        self.top5_accuracy = MulticlassAccuracy(
            num_classes=n_classes, top_k=5, average="micro"
        )

    def forward(self, batch):
        z = self.transformer(batch)
        mlm_logits = self.output_head(z[:, 1:]).squeeze(-1)

        clf_logits = self.transformer.output_head(z[:, 0, :])
        return mlm_logits, clf_logits


    def training_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        batch = self.shuffler(batch)

        mlm_logits, clf_logits = self(batch)

        mlm_logits_train = self.train_indices_select(mlm_logits, batch["train_indices"])
        trues_train = self.train_indices_select(
            batch["intensity_true"], batch["train_indices"]
        )
        mlm_loss = F.binary_cross_entropy_with_logits(
            mlm_logits_train, trues_train.to(self.dtype)
        )

        indexer = batch["species"] < self.n_species
        clf_loss = F.cross_entropy(clf_logits[indexer], batch["species"][indexer])

        self.log(
            "train_loss", mlm_loss + clf_loss * (1 if self.clf else 0), batch_size=len(batch["intensity"])
        )
        return mlm_loss + clf_loss * (
            self.lmbda if ((torch.rand(1) < self.clf_train_p).item() and self.clf) else 0
        )

    def validation_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        batch = self.shuffler(batch)

        mlm_logits, clf_logits = self(batch)

        mlm_logits_train = self.train_indices_select(mlm_logits, batch["train_indices"])
        trues_train = self.train_indices_select(
            batch["intensity_true"], batch["train_indices"]
        )
        mlm_loss = F.binary_cross_entropy_with_logits(
            mlm_logits_train, trues_train.to(self.dtype)
        )

        self.log(
            "val_mlmloss",
            mlm_loss,
            on_step=False,
            on_epoch=True,
            batch_size=len(mlm_logits_train),
            sync_dist=True,
        )

        self.auroc(mlm_logits_train, trues_train)
        self.log(
            "val_mlmmicro_auc",
            self.auroc,
            on_step=False,
            on_epoch=True,
            batch_size=len(mlm_logits_train),
            sync_dist=True,
        )

        if self.clf:
            indexer = batch["species"] < self.n_species
            if indexer.sum() > 0:
                clf_loss = F.cross_entropy(
                    clf_logits[indexer], batch["species"][indexer]
                )

                self.log(
                    "val_clfloss",
                    clf_loss,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )
                self.accuracy(clf_logits[indexer], batch["species"][indexer])
                self.log(
                    "val_clfacc",
                    self.accuracy,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )
                self.top5_accuracy(clf_logits[indexer], batch["species"][indexer])
                self.log(
                    "val_clftop5_acc",
                    self.top5_accuracy,
                    on_step=False,
                    on_epoch=True,
                    batch_size=indexer.sum(),
                    sync_dist=True,
                )

    def predict_step(self, batch, batch_idx):
        batch["intensity"] = batch["intensity"].to(self.dtype)
        batch["mz"] = batch["mz"].to(self.dtype)

        batch = self.shuffler(batch)

        mlm_logits, clf_logits = self(batch)

        logits_train = self.train_indices_select(mlm_logits, batch["train_indices"])
        trues_train = self.train_indices_select(
            batch["intensity_true"], batch["train_indices"]
        )
        mzs_train = self.train_indices_select(batch["mz"], batch["train_indices"])
        intensities_train = self.train_indices_select(
            batch["intensity"], batch["train_indices"]
        )

        if self.clf:
            return (
                np.array(batch["loc"])[
                    torch.where(batch["train_indices"])[0].cpu().numpy()
                ],
                batch["species"][torch.where(batch["train_indices"])[0]],
                mzs_train,
                intensities_train,
                logits_train,
                trues_train,
                clf_logits,
                batch["species"],
            )
        else:
            return (
                np.array(batch["loc"])[
                    torch.where(batch["train_indices"])[0].cpu().numpy()
                ],
                batch["species"][torch.where(batch["train_indices"])[0]],
                mzs_train,
                intensities_train,
                logits_train,
                trues_train,
            )

    def shuffler(self, batch):
        mz = batch["mz"]
        intensity = batch["intensity"]
        n = len(mz) // 2

        all_indices = torch.stack(torch.where(mz)).T

        shuff, pos = torch.chunk(
            torch.randperm(len(all_indices), device=all_indices.device)[
                : int(len(all_indices) * self.p)
            ],
            2,
        )
        shuffled_shuff = shuff[torch.randperm(len(shuff), device=shuff.device)]

        indexer = (all_indices[shuff] != all_indices[shuffled_shuff])[:, 0]

        shuff = shuff[indexer]
        shuffled_shuff = shuffled_shuff[indexer]
        pos = pos[indexer]

        train_indices = torch.zeros_like(mz.bool())
        train_indices[all_indices[pos][:, 0], all_indices[pos][:, 1]] = True
        train_indices[all_indices[shuff][:, 0], all_indices[shuff][:, 1]] = True

        intensity_true = torch.ones_like(mz).long()
        intensity_true[all_indices[shuff][:, 0], all_indices[shuff][:, 1]] = 0

        all_indices[shuff] = all_indices[shuffled_shuff]

        batch["mz"] = mz[all_indices[:, 0], all_indices[:, 1]].view_as(mz)
        batch["intensity"] = intensity[all_indices[:, 0], all_indices[:, 1]].view_as(
            intensity
        )
        batch["intensity_true"] = intensity_true
        batch["train_indices"] = train_indices
        return batch



import os

os.environ["OMP_NUM_THREADS"] = "4"  # export OMP_NUM_THREADS=1
os.environ["OPENBLAS_NUM_THREADS"] = "4"  # export OPENBLAS_NUM_THREADS=1
os.environ["MKL_NUM_THREADS"] = "4"  # export MKL_NUM_THREADS=1
os.environ["VECLIB_MAXIMUM_THREADS"] = "4"  # export VECLIB_MAXIMUM_THREADS=1
os.environ["NUMEXPR_NUM_THREADS"] = "4"  # export NUMEXPR_NUM_THREADS=1

import numpy as np
import h5torch
import torch
import pandas as pd
from sklearn.metrics import roc_auc_score
import torch.nn.functional as F
from lightning.pytorch import Trainer
from torch.utils.data import TensorDataset, DataLoader
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping
import maldi_nn.nn as maldinn
from maldi_nn.models import MaldiLightningModule
from torchmetrics.classification import BinaryAUROC
import argparse
from sklearn.model_selection import ParameterGrid
from xgboost import XGBClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
import ast


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass


class BaselineModel(MaldiLightningModule):
    def __init__(
        self,
        spectrum_kwargs={},
        weight_decay=0,
        lr=1e-4,
        lr_decay_factor=1.00,
        warmup_steps=250,
    ):
        super().__init__(
            lr=lr,
            weight_decay=weight_decay,
            lr_decay_factor=lr_decay_factor,
            warmup_steps=warmup_steps,
        )
        self.save_hyperparameters()

        self.spectrum_embedder = maldinn.MLP(**spectrum_kwargs)
        self.auroc = BinaryAUROC()

    def forward(self, batch):
        return self.spectrum_embedder(batch).squeeze(-1)

    def training_step(self, batch, batch_idx):
        X, y = batch
        batch = {"intensity": X.to(self.dtype)}

        logits = self(batch)

        loss = F.binary_cross_entropy_with_logits(logits, y)

        self.log("train_loss", loss, batch_size=len(logits))
        return loss

    def validation_step(self, batch, batch_idx):
        X, y = batch
        batch = {"intensity": X.to(self.dtype)}

        logits = self(batch)

        loss = F.binary_cross_entropy_with_logits(logits, y)

        self.log("val_loss", loss, batch_size=len(logits))

        self.auroc(logits, y)
        self.log(
            "val_auroc",
            self.auroc,
            on_step=False,
            on_epoch=True,
            batch_size=len(logits),
        )


dict_ = {
    "S": [256, 128],
    "M": [512, 256, 128],
    "L": [1024, 512, 256, 128],
    "XL": [2048, 1024, 512, 256, 128],
}


def main_MLP(args):
    f = h5torch.File(args.path, "r")

    dr = f["central/indices"][1]
    sp = f["0/species"][:][f["central/indices"][0]]
    comb, cnt = np.unique(
        (pd.Series(sp.astype(str)) + "_" + pd.Series(dr.astype(str))).values[
            (f["unstructured/split"][:] == b"A_train")
        ],
        return_counts=True,
    )
    asrt = np.argsort(cnt)

    locs = []
    drug_names = []
    preds = []
    trues = []

    for t in comb[asrt][-300:][::-1]:
        print(t)
        s, d = t.split("_")
        col_of_drug = int(d)
        rows_with_species = np.where(f["0/species"][:] == int(s))[0]

        indices = np.logical_and(
            f["central/indices"][:][1] == col_of_drug,
            np.isin(f["central/indices"][:][0], rows_with_species),
        )

        train = np.logical_and(indices, f["unstructured/split"][:] == b"A_train")
        val = np.logical_and(indices, f["unstructured/split"][:] == b"A_val")
        test = np.logical_and(indices, f["unstructured/split"][:] == b"A_test")

        X_train = f["0/intensity"][f["central/indices"][:][0][train]]
        y_train = (f["central/data"][:][train] != b"S").astype(int)

        X_val = f["0/intensity"][f["central/indices"][:][0][val]]
        y_val = (f["central/data"][:][val] != b"S").astype(int)

        X_test = f["0/intensity"][f["central/indices"][:][0][test]]
        y_test = (f["central/data"][:][test] != b"S").astype(int)
        if (
            (len(np.unique(y_test)) <= 1)
            or (len(np.unique(y_val)) <= 1)
            or (len(np.unique(y_train)) <= 1)
        ):
            continue

        kwargs = {
            "n_inputs": 6000,
            "n_outputs": 1,
            "layer_dims": dict_[args.mlp_size],
            "layer_or_batchnorm": "layer",
            "dropout": 0.2,
        }

        valscores_all = []
        preds_test_all = []
        for lr in [1e-4, 5e-4, 1e-3, 5e-3]:
            model = BaselineModel(
                spectrum_kwargs=kwargs,
                lr=lr,
                weight_decay=0,
                lr_decay_factor=1.00,
                warmup_steps=250,
            )

            train_set = TensorDataset(
                torch.tensor(X_train).float(), torch.tensor(y_train).float()
            )
            val_set = TensorDataset(
                torch.tensor(X_val).float(), torch.tensor(y_val).float()
            )
            test_set = TensorDataset(
                torch.tensor(X_test).float(), torch.tensor(y_test).float()
            )
            train_dl = DataLoader(
                train_set, batch_size=128, shuffle=True, pin_memory=True
            )
            val_dl = DataLoader(val_set, batch_size=128, shuffle=False, pin_memory=True)
            test_dl = DataLoader(
                test_set, batch_size=128, shuffle=False, pin_memory=True
            )

            val_ckpt = ModelCheckpoint(monitor="val_auroc", mode="max")
            callbacks = [val_ckpt]
            callbacks += [EarlyStopping(monitor="val_auroc", patience=10, mode="max")]

            trainer = Trainer(
                accelerator="gpu",
                devices=args.mlp_devices,
                max_epochs=250,
                callbacks=callbacks,
            )
            trainer.fit(model, train_dl, val_dl)

            model = BaselineModel.load_from_checkpoint(val_ckpt.best_model_path)
            model = model.eval().cpu()
            with torch.no_grad():
                preds_val = []
                for i in np.arange(0, len(val_set), 128):
                    X, _ = val_set[i : i + 128]
                    batch = {
                        "intensity": X.to(model.dtype),
                        "mz": (torch.arange(2000, 20000, 3) + 1.5)
                        .expand(X.shape[0], -1)
                        .to(model.dtype),
                    }
                    preds_val.append(model(batch).numpy())

                valscores_all.append(roc_auc_score(y_val, np.concatenate(preds_val)))

                preds_test = []
                for i in np.arange(0, len(test_set), 128):
                    X, _ = test_set[i : i + 128]
                    batch = {
                        "intensity": X.to(model.dtype),
                        "mz": (torch.arange(2000, 20000, 3) + 1.5)
                        .expand(X.shape[0], -1)
                        .to(model.dtype),
                    }
                    preds_test.append(model(batch).numpy())
                preds_test_all.append(np.concatenate(preds_test))

        # final model
        max_index = np.argmax(valscores_all)

        loc_ = f["0/loc"][:][f["central/indices"][:][0][test]]
        drug_name = f["1/drug_names"][:][f["central/indices"][:][1][test]]

        preds.append(preds_test_all[max_index])
        trues.append(y_test)
        locs.append(loc_)
        drug_names.append(drug_name)

    np.savez(
        args.outputs,
        **{
            "trues": np.concatenate(trues),
            "preds": np.concatenate(preds),
            "locs": np.concatenate(locs),
            "drug_names": np.concatenate(drug_names),
        }
    )

def main_lr(args):
    f = h5torch.File(args.path, "r")

    dr = f["central/indices"][1]
    sp = f["0/species"][:][f["central/indices"][0]]
    comb, cnt = np.unique(
        (pd.Series(sp.astype(str)) + "_" + pd.Series(dr.astype(str))).values[
            (f["unstructured/split"][:] == b"A_train")
        ],
        return_counts=True,
    )
    asrt = np.argsort(cnt)

    locs = []
    drug_names = []
    preds = []
    trues = []

    for t in comb[asrt][-300:][::-1]:
        print(t)
        s, d = t.split("_")
        col_of_drug = int(d)
        rows_with_species = np.where(f["0/species"][:] == int(s))[0]

        indices = np.logical_and(
            f["central/indices"][:][1] == col_of_drug,
            np.isin(f["central/indices"][:][0], rows_with_species),
        )

        train = np.logical_and(indices, f["unstructured/split"][:] == b"A_train")
        val = np.logical_and(indices, f["unstructured/split"][:] == b"A_val")
        test = np.logical_and(indices, f["unstructured/split"][:] == b"A_test")

        X_train = f["0/intensity"][f["central/indices"][:][0][train]]
        y_train = (f["central/data"][:][train] != b"S").astype(int)

        X_val = f["0/intensity"][f["central/indices"][:][0][val]]
        y_val = (f["central/data"][:][val] != b"S").astype(int)

        X_test = f["0/intensity"][f["central/indices"][:][0][test]]
        y_test = (f["central/data"][:][test] != b"S").astype(int)
        if (
            (len(np.unique(y_test)) <= 1)
            or (len(np.unique(y_val)) <= 1)
            or (len(np.unique(y_train)) <= 1)
        ):
            continue

        # tune
        lr_grid = ParameterGrid(
            {
                "norm": ["passthrough", "standardscaler"], 
                "penalty": ["l2"],
                "C": 10.0 ** np.arange(-3, 4),
            }
        )

        scores = []
        for params in lr_grid:
            if params["norm"] == "standardscaler":
                model = Pipeline(
                    steps=[
                        ("norm", StandardScaler()),
                        (
                            "lr",
                            LogisticRegression(
                                solver="lbfgs",
                                max_iter=500,
                                penalty=params["penalty"],
                                C=params["C"],
                            ),
                        ),
                    ]
                )
            else:
                model = LogisticRegression(
                    solver="lbfgs",
                    max_iter=500,
                    penalty=params["penalty"],
                    C=params["C"],
                )
            model.fit(X_train, y_train)
            print(params, roc_auc_score(y_val, model.predict_proba(X_val)[:, 1]))
            scores.append(
                [params, "lr", roc_auc_score(y_val, model.predict_proba(X_val)[:, 1])]
            )

        # final model
        max_index = np.argmax([j[-1] for j in scores])

        if scores[max_index][0]["norm"] == "standardscaler":
            m = Pipeline(
                steps=[
                    ("norm", StandardScaler()),
                    (
                        "lr",
                        LogisticRegression(
                            solver="lbfgs",
                            max_iter=500,
                            penalty=scores[max_index][0]["penalty"],
                            C=scores[max_index][0]["C"],
                        ),
                    ),
                ]
            )
        else:
            m = LogisticRegression(
                solver="lbfgs",
                max_iter=500,
                penalty=scores[max_index][0]["penalty"],
                C=scores[max_index][0]["C"],
            )

        m.fit(X_train, y_train)

        loc_ = f["0/loc"][:][f["central/indices"][:][0][test]]
        drug_name = f["1/drug_names"][:][f["central/indices"][:][1][test]]

        preds.append(m.predict_proba(X_test)[:, 1])
        trues.append(y_test)
        locs.append(loc_)
        drug_names.append(drug_name)

    np.savez(
        args.outputs,
        **{
            "trues": np.concatenate(trues),
            "preds": np.concatenate(preds),
            "locs": np.concatenate(locs),
            "drug_names": np.concatenate(drug_names),
        }
    )

def main_xgb(args):
    f = h5torch.File(args.path, "r")

    dr = f["central/indices"][1]
    sp = f["0/species"][:][f["central/indices"][0]]
    comb, cnt = np.unique(
        (pd.Series(sp.astype(str)) + "_" + pd.Series(dr.astype(str))).values[
            (f["unstructured/split"][:] == b"A_train")
        ],
        return_counts=True,
    )
    asrt = np.argsort(cnt)

    locs = []
    drug_names = []
    preds = []
    trues = []

    for t in comb[asrt][-300:][::-1]:
        print(t)
        s, d = t.split("_")
        col_of_drug = int(d)
        rows_with_species = np.where(f["0/species"][:] == int(s))[0]

        indices = np.logical_and(
            f["central/indices"][:][1] == col_of_drug,
            np.isin(f["central/indices"][:][0], rows_with_species),
        )

        train = np.logical_and(indices, f["unstructured/split"][:] == b"A_train")
        val = np.logical_and(indices, f["unstructured/split"][:] == b"A_val")
        test = np.logical_and(indices, f["unstructured/split"][:] == b"A_test")

        X_train = f["0/intensity"][f["central/indices"][:][0][train]]
        y_train = (f["central/data"][:][train] != b"S").astype(int)

        X_val = f["0/intensity"][f["central/indices"][:][0][val]]
        y_val = (f["central/data"][:][val] != b"S").astype(int)

        X_test = f["0/intensity"][f["central/indices"][:][0][test]]
        y_test = (f["central/data"][:][test] != b"S").astype(int)
        if (
            (len(np.unique(y_test)) <= 1)
            or (len(np.unique(y_val)) <= 1)
            or (len(np.unique(y_train)) <= 1)
        ):
            continue

        # tune
        xgb_grid = ParameterGrid(
            {
                "n_estimators": [25, 50, 100, 200],
                "learning_rate": 10.0 ** np.arange(-3, 1),
            }
        )

        scores = []
        for params in xgb_grid:
            model = XGBClassifier(**params)

            model.fit(X_train, y_train)
            print(params, roc_auc_score(y_val, model.predict_proba(X_val)[:, 1]))
            scores.append(
                [params, "xgb", roc_auc_score(y_val, model.predict_proba(X_val)[:, 1])]
            )

        # final model
        max_index = np.argmax([j[-1] for j in scores])

        m = XGBClassifier(**scores[max_index][0])
        m.fit(X_train, y_train)

        loc_ = f["0/loc"][:][f["central/indices"][:][0][test]]
        drug_name = f["1/drug_names"][:][f["central/indices"][:][1][test]]

        preds.append(m.predict_proba(X_test)[:, 1])
        trues.append(y_test)
        locs.append(loc_)
        drug_names.append(drug_name)

    np.savez(
        args.outputs,
        **{
            "trues": np.concatenate(trues),
            "preds": np.concatenate(preds),
            "locs": np.concatenate(locs),
            "drug_names": np.concatenate(drug_names),
        }
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Training script for non-recommender AMR baselines.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument(
        "outputs",
        type=str,
        metavar="outputs.npz",
        help="numpy .npz file to write (test) predictions into.",
    )
    parser.add_argument(
        "modeltype",
        type=str,
        metavar="modeltype",
        choices=["MLP","lr", "xgb"],
        help="Which modeltype to use as baseline, choices: {%(choices)s}",
    )

    parser.add_argument(
        "--mlp_size",
        type=str,
        choices=["S", "M", "L", "XL", "Linear"],
        default=["M"],
        help="Which size spectrum embedder to use for MLP, choices: {%(choices)s}",
    )
    parser.add_argument(
        "--mlp_devices",
        type=ast.literal_eval,
        default=1,
        help="devices to use for MLP. Input an integer to specify a number of gpus or a list e.g. [1] or [0,1,3] to specify which gpus.",
    )
    args = parser.parse_args()

    if args.modeltype == "MLP":
        main_MLP(args)
    elif args.modeltype == "lr":
        main_lr(args)
    elif args.modeltype == "xgb":
        main_xgb(args)
    